
from enum import StrEnum, IntEnum

LangMap = {
    "zh-CN": "zh-cn",
    "zh-TW": "zh-tw",
    "de": "de-de",
    "en": "en-us",
    "es": "es-es",
    "fr": "fr-fr",
    "id": "id-id",
    "it": "it-it",
    "ja": "ja-jp",
    "ko": "ko-kr",
    "pt": "pt-pt",
    "ru": "ru-ru",
    "th": "th-th",
    "vi": "vi-vn",
    "tr": "tr",
}

class MiHoMoLink(StrEnum):
    PARSED = "https://api.mihomo.me/sr_info_parsed/{uid}"
    NAKED = "https://api.mihomo.me/sr_info/{uid}"
    
class EnkaLink(StrEnum):
    ENKA = "https://enka.network/api/hsr/uid/{uid}"
    HASH = "https://enka.network/api/profile/{name}/hoyos/"
    BUILD = "https://enka.network/api/profile/{name}/hoyos/{hash}/builds/"

class ApiType(IntEnum):
    MiHoMo = 1
    Enka = 2
    HoYoLab = 3

class LangMiHoMo(StrEnum):
    EN = "en"
    EN = "en"
    CHT = "cht"
    CN = "cn"
    DE = "de"
    ES = "es"
    FR = "fr"
    ID = "id"
    JP = "jp"
    KR = "kr"
    PT = "pt"
    RU = "ru"
    TH = "th"
    VI = "vi"
    UA = "ua"


import aiohttp
from typing import Optional, Union
from .api import ApiMiHoMo
from ..model import api_mihomo
from ..tools import http, translator, options, ukrainization
from .error import StarRailCardError
from .hoyolab_parsed import AssetHoYoLabParsed

    
LANG_MAP = {
    "cn": "zh-cn",
    "cht": "zh-tw",
    "de": "de-de",
    "en": "en-us",
    "es": "es-es",
    "fr": "fr-fr",
    "id": "id-id",
    "it": "it-it",
    "ja": "ja-jp",
    "ko": "ko-kr",
    "pt": "pt-pt",
    "ru": "ru-ru",
    "th": "th-th",
    "vi": "vi-vn",
    "tr": "tr",
}

class HoYoLabApi:
    """Class for interacting with the MiHoMo API."""

    def __init__(self, uid: str = 0, lang: str = "en"):
        """Initialize the HoYoLabApi object."""
        self.uid: str = uid
        self.lang: str = translator.SUPPORTED_LANGUAGES.get(lang, "en")
        self.ua_lang: str = lang == "en"
        api_mihomo.UA_LANG = False
        if lang == "ua":
            self.ua_lang = True
            api_mihomo.UA_LANG = True
            self.lang = "en"
        
        self.convert_lang = LANG_MAP.get(self.lang)



    async def get(self,cookie: dict, uid: Union[str,int] = 0):
        try:
            import genshin
        except ImportError:
            raise StarRailCardError(100, "Install the genshin.py module\n- pip install genshin")
        
        if uid == 0:
            uid = self.uid
            if uid == 0:
                raise StarRailCardError(5, "Specify UID")
            
        player = await ApiMiHoMo(uid, self.lang).get(parse= False)
        
        try:
            client = genshin.Client(cookie, game= genshin.Game.STARRAIL, lang= self.convert_lang)
            data = await client.get_starrail_characters(uid)
        except Exception as e:
            print(e)
            raise
        data = await AssetHoYoLabParsed(data, data.property_info).collect()
        
        
        if self.ua_lang:
            await ukrainization.TranslateDataManager().load_translate_data()

        return api_mihomo.MiHoMoApi(player=player["player"], characters= data, dont_update_link = False)
    

    
    
        


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

from pydantic import BaseModel, Field
from typing import List, Optional,Final, Union

from ..tools.ukrainization import TranslateDataManager

UA_LANG = False
MAIN_LINK: Final[str] = "https://raw.githubusercontent.com/Mar-7th/StarRailRes/master/{icon}"


def hex_to_rgba(hex_code):
    hex_code = hex_code.strip('#')
    
    red = int(hex_code[0:2], 16)
    green = int(hex_code[2:4], 16)
    blue = int(hex_code[4:6], 16)
    
    alpha = 255 
    
    return red, green, blue, alpha

class MemoryInfo(BaseModel):
    chaos_id: Optional[int]
    chaos_level: Optional[int]
    abyss_level: Optional[int] =  Field(0, alias= "abyssLevel")
    abyss_star_count: Optional[int] =  Field(0, alias= "abyssStarCount")
    
    class Config:
        populate_by_name = True
    
class SpaceInfo(BaseModel):
    relic_count: Optional[int] =  Field(0, alias= "relicCount")
    music_count: Optional[int] =  Field(0, alias= "musicCount")
    book_count: Optional[int] =  Field(0, alias= "bookCount")
    memory_data: Optional[MemoryInfo]
    universe_level: Optional[int]
    light_cone_count: Optional[int]
    avatar_count: Optional[int]
    achievement_count: Optional[int]

    class Config:
        populate_by_name = True
        
class Avatar(BaseModel):
    id: str
    name: str
    icon: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.icon == "" or self.icon is None:
            self.icon = "icon/avatar/202018.png"
        
        if UA_LANG:
            self.name = TranslateDataManager._data.avatar.get(self.id, self.name)
        
class Player(BaseModel):
    uid: Optional[Union[str,int]]
    nickname: Optional[str]
    level: int
    world_level: int
    friend_count: int
    avatar: Avatar
    signature: Optional[str]
    is_display: Optional[bool]
    space_info: Optional[SpaceInfo]

class RelicMainAffix(BaseModel):
    type: Optional[str]
    field: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    value: float
    display: Optional[str]
    percent: bool
    
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
        if UA_LANG:
            self.name = TranslateDataManager._data.stats.get(self.type, self.name)

class RelicSubAffix(BaseModel):
    type: Optional[str]
    field: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    value: float
    display: Optional[str]
    percent: bool
    count: int
    step: int

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
        if UA_LANG:
            self.name = TranslateDataManager._data.stats.get(self.type, self.name)
            
class Relic(BaseModel):
    id: Optional[str]
    name: Optional[str]
    set_id: Optional[str]
    set_name: Optional[str]
    rarity: int
    level: int
    icon: Optional[str]
    main_affix: RelicMainAffix
    sub_affix: List[RelicSubAffix]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
        if UA_LANG:
            self.set_name = TranslateDataManager._data.relict_sets.get(self.set_id, self.set_name)
            
class RelicSetProperties(BaseModel):
    type: Optional[str]
    field: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    value: float
    display: Optional[str]
    percent: bool
    
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
        if UA_LANG:
            self.name = TranslateDataManager._data.stats.get(self.type, self.name)

class RelicSet(BaseModel):
    id: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    num: int
    desc: Optional[str]
    properties: List[RelicSetProperties] = Field([], alias="properties")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
            
        if UA_LANG:
            self.name = TranslateDataManager._data.relict_sets.get(self.id, self.name)
            
class LightConeAttributes(BaseModel):
    field: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    value: float
    display: Optional[str]
    percent: bool
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if UA_LANG:
            self.name = TranslateDataManager._data.stats.get(self.field, self.name)

class Path(BaseModel):
    id: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)    
        if UA_LANG:
            self.name = TranslateDataManager._data.paths.get(self.id, self.name)

class LightCone(BaseModel):
    id: Optional[str]
    name: Optional[str]
    rarity: int
    rank: int
    level: int
    promotion: Optional[int] = 0
    icon: Optional[str]
    preview: Optional[str]
    portrait: Optional[str]
    path: Path
    attributes: List[LightConeAttributes]
    properties: Optional[List[LightConeAttributes]]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if UA_LANG:
            self.name = TranslateDataManager._data.weapons.get(self.id, self.name)
            
class ColorElement(BaseModel):
    hex: Optional[str]
    rgba: Optional[tuple]
     
class Element(BaseModel):
    id: Optional[str]
    name: Optional[str]
    color: Optional[str]
    icon: Optional[str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.color = ColorElement(hex = self.color, rgba = hex_to_rgba(self.color))
        
        if UA_LANG:
            self.name = TranslateDataManager._data.element.get(self.id, self.name)
               
class Skill(BaseModel):
    id: Optional[str]
    name: Optional[str]
    level: int
    max_level: int
    element: Optional[Element]
    type: Optional[str]
    type_text: Optional[str]
    effect: Optional[str]
    effect_text: Optional[str]
    simple_desc: Optional[str]
    desc: Optional[str]
    icon: Optional[str]
    
class SkillTree(BaseModel):
    id: Optional[str]
    level: int
    anchor: Optional[str]
    max_level: int
    icon: Optional[str]
    parent: Optional[str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
class CharacterAttributes(BaseModel):
    field: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    value: float
    display: Optional[str]
    percent: bool
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if UA_LANG:
            self.name = TranslateDataManager._data.stats.get(self.field, self.name)
            
class CharacterProperties(BaseModel):
    type: Optional[str]
    field: Optional[str]
    name: Optional[str]
    icon: Optional[str]
    value: float
    display: Optional[str]
    percent: bool
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if UA_LANG:
            self.name = TranslateDataManager._data.stats.get(self.field, self.name)            

  
class Character(BaseModel):
    id: Optional[str]
    name: Optional[str]
    rarity: int
    rank: int
    level: int
    promotion: int
    icon: Optional[str]
    preview: Optional[str]
    portrait: Optional[str]
    rank_icons: list
    path: Path
    element: Optional[Element]
    skills: List[Skill]
    skill_trees: List[SkillTree]
    light_cone: Optional[LightCone]
    relics: Optional[List[Relic]]
    relic_sets: Optional[List[RelicSet]] = Field([], alias="relic_sets")
    attributes: List[CharacterAttributes]
    additions: List[CharacterAttributes]
    properties: Optional[List[CharacterProperties]]
    build: Optional[dict] = Field(None)
    pos: list
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)        
        if UA_LANG:
            self.name = TranslateDataManager._data.avatar.get(self.id, self.name)
    
class MiHoMoApi(BaseModel):
    player: Player
    characters: List[Character]
    dont_update_link: Optional[bool] = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.dont_update_link:
            self.player.avatar.icon = MAIN_LINK.format(icon = self.player.avatar.icon)
            for character in self.characters:
                character.icon = MAIN_LINK.format(icon=character.icon)
                character.preview = MAIN_LINK.format(icon=character.preview)
                character.portrait = MAIN_LINK.format(icon=character.portrait)
                character.rank_icons = [MAIN_LINK.format(icon=icon) for icon in character.rank_icons]

                if character.light_cone:
                    character.light_cone.icon = MAIN_LINK.format(icon=character.light_cone.icon)
                    character.light_cone.preview = MAIN_LINK.format(icon=character.light_cone.preview)
                    character.light_cone.portrait = MAIN_LINK.format(icon=character.light_cone.portrait)

                for relic in character.relics:
                    relic.icon = MAIN_LINK.format(icon=relic.icon)
                    for sub_affix in relic.sub_affix:
                        sub_affix.icon = MAIN_LINK.format(icon=sub_affix.icon)
                    relic.main_affix.icon = MAIN_LINK.format(icon=relic.main_affix.icon)

                for relic_set in character.relic_sets:
                    relic_set.icon = MAIN_LINK.format(icon=relic_set.icon)
                    for property in relic_set.properties:
                        property.icon = MAIN_LINK.format(icon=property.icon)

                for skill in character.skills:
                    if skill.icon:
                        skill.icon = MAIN_LINK.format(icon=skill.icon)

                for skill_tree in character.skill_trees:
                    if skill_tree.icon:
                        skill_tree.icon = MAIN_LINK.format(icon=skill_tree.icon)

                for attribute in character.attributes:
                    attribute.icon = MAIN_LINK.format(icon=attribute.icon)

                
                for addition in character.additions:
                    addition.icon = MAIN_LINK.format(icon=addition.icon)

                if character.properties:
                    for property in character.properties:
                        property.icon = MAIN_LINK.format(icon=property.icon)

                character.element.icon = MAIN_LINK.format(icon = character.element.icon)

                for skills in character.skills:
                    if "icon" in skills:
                        skills = MAIN_LINK.format(icon = skills)

                character.path.icon = MAIN_LINK.format(icon=character.path.icon)

                
                
                if not character.light_cone is None:
                    for light_cone_attr in character.light_cone.attributes:
                        light_cone_attr.icon = MAIN_LINK.format(icon=light_cone_attr.icon)
                        
                    if character.light_cone.properties:
                        for property in character.light_cone.properties:
                            property.icon = MAIN_LINK.format(icon=property.icon)




class ApiError(Exception):
    """Base exception for errors when working with the MiHoMo API."""

    def __init__(self, code: int, message: str, status: int = 400) -> None:
        """Initialize the error with a code and message."""
        super().__init__(f"[{code}] {message} Status: {status}")
        self.code = code
        self.message = message
        self.status = status

class StarRailCardError(ApiError):
    """Exception specific to MiHoMo errors."""
    pass


import aiohttp
from typing import Optional, Union
import os

    
from ..tools import http, translator, ukrainization
from .enka_parsed import *
from .error import StarRailCardError
from ..tools.json_data import JsonManager
from ..tools.enums import PathData
from ..tools.translator import SUPPORTED_LANGUAGES
from .enka_parsed import AssetEnkaParsed
from ..model import api_mihomo

_API_ENKA = "https://enka.network/api/hsr/uid/{uid}"
_API_ENKA_HASH = "https://enka.network/api/profile/{name}/hoyos/"
_API_ENKA_BUILD = "https://enka.network/api/profile/{name}/hoyos/{hash}/builds/"


_INDEX_MIHOMO = "https://raw.githubusercontent.com/Mar-7th/StarRailRes/master/index_new/{lang}/{index}.json"

_INDEX_NAME = [
    "avatars",
    "character_promotions",
    "character_ranks",
    "character_skill_trees",
    "character_skills",
    "characters",
    "elements",
    "light_cone_promotions",
    "light_cone_ranks",
    "light_cones",
    "paths",
    "properties",
    "relic_main_affixes",
    "relic_sets",
    "relic_sub_affixes",
    "relics"    
]



class ApiEnkaNetwork:
    def __init__(self, uid: Union[int, str] = 0, lang: str = "en", parsed: bool = True) -> None:
        self.uid: str = uid
        self.parsed: bool = parsed
        self.lang: str = translator.SUPPORTED_LANGUAGES.get(lang, "en")
        self.ua_lang: str = lang == "en"
        api_mihomo.UA_LANG = False
        if lang == "ua":
            self.ua_lang = True
            api_mihomo.UA_LANG = True
    
    async def get_build(self, name: str, hash: str, uid: Union[int, str]  = 0):
        
        if uid == 0:
            uid = self.uid
         
        if uid == 0:
            raise StarRailCardError(5, "Specify UID")
            
        try:
            data_build = await http.AioSession.get(_API_ENKA_BUILD.format(name=name, hash = hash))
            data = await self.get(uid, parsed = False)
            data["detailInfo"]["avatarDetailList"] = [data_build]
            if self.parsed:
                data = await AssetEnkaParsed(data).collect(True)
            else:
                return data
        except aiohttp.ClientConnectionError:
            raise StarRailCardError(1, "Server is not responding")
        except aiohttp.ClientResponseError as e:
            raise StarRailCardError(e.status, f"Server returned status code {e.status}")
        
        if self.ua_lang:
            await ukrainization.TranslateDataManager().load_translate_data()

                
        return api_mihomo.MiHoMoApi(player=data["player"], characters=data["characters"], dont_update_link = False)
    
    
    async def get_hash(self, name: str, hash: bool = True):
        """Return user's HASH
        Args:
            name (str): Profile nickname on the EnkaNetwork website
            hash (bool, optional): Return only the user's hash
        Raises:
            StarRailCardError: Server is not responding
        Returns:
            list: List of included dictionaries with hash and profile information
        """
        try:
            data = await http.AioSession.get(_API_ENKA_HASH.format(name=name))
        except aiohttp.ClientConnectionError:
            raise StarRailCardError(1, "Server is not responding")
        except aiohttp.ClientResponseError as e:
            raise StarRailCardError(e.status, f"Server returned status code {e.status}")
        
        if hash:
            return [
                    {
                        "enka": name,
                        "hash": key,
                        "region": data[key]["region"],
                        "uid": data[key]["uid"],
                        "icon": data[key]["player_info"]["headIcon"],
                        "name": data[key]["player_info"]["nickname"],
                        "level": data[key]["player_info"]["level"]
                    } for key in data if data[key]["hoyo_type"] == 1
                ]
        
        return data
            
    async def get(self, uid: Union[str,int] = 0, parsed: bool = True):
        """Get data from the EnkaNetwork API."""
        
        if uid == 0:
            uid = self.uid
            
            if uid == 0:
                raise StarRailCardError(5, "Specify UID")
        
        try:
            data = await http.AioSession.get(_API_ENKA.format(uid=uid))

            if self.parsed and parsed:
                data = await AssetEnkaParsed(data).collect()
            else:
                return data
        except aiohttp.ClientConnectionError:
            raise StarRailCardError(1, "Server is not responding")
        except aiohttp.ClientResponseError as e:
            raise StarRailCardError(e.status, f"Server returned status code {e.status}")
        
        if self.ua_lang:
            await ukrainization.TranslateDataManager().load_translate_data()

        return api_mihomo.MiHoMoApi(player=data["player"], characters=data["characters"], dont_update_link = False)
        
    async def update_assets(self, lang = None):
        print("===START UPDATE ASSET===")
        for name in _INDEX_NAME:
            for langs in SUPPORTED_LANGUAGES:
                if langs == "ua":
                    continue
                if not lang is None:
                    if langs != lang:
                        continue
                data = await http.AioSession.get(_INDEX_MIHOMO.format(lang = langs, index=name))
                if not os.path.exists(PathData.ENKA_INDEX.value / langs):
                    os.makedirs(PathData.ENKA_INDEX.value / langs)
                await JsonManager(PathData.ENKA_INDEX.value / langs /f"{name}.json").write(data)
                
            print(f"- Updated file: {name}")
        print("===END UPDATE ASSET===")        
        


import aiohttp
from typing import Union
from ..model import api_mihomo
from .error import StarRailCardError
from ..tools import http, translator, ukrainization, options
from .config import (ApiType,
                     LangMiHoMo,
                     EnkaLink,
                     MiHoMoLink,
                     LangMap
                     )

class ShowcaseApi:
    def __init__(self, lang: LangMiHoMo = LangMiHoMo.EN, user_agent: str = None,  proxy: Union[dict,str] = None, parsed_info: bool = True,
                 force_update: bool = True, v: int = 2) -> None:
        self.parsed_info = parsed_info
        self.lang = translator.SUPPORTED_LANGUAGES.get(lang, "en")
        self.force_update = force_update
        self.proxy = proxy
        self.headers = {
            "User-Agent": options.get_user_agent(user_agent)
        }
        api_mihomo.UA_LANG = False
        if lang == "ua":
            self.lang = "en"
            self.ua_lang = True
            api_mihomo.UA_LANG = True
        
        self.convert_lang = LangMap.get(self.lang)
        self.v = v
    
    async def get_params(self, api_id: int = 1):
        if api_id == 2:
            return {}
        return {
                'lang': self.lang,
                'is_force_update': str(self.force_update),
                'version': f"v{self.v}"
            }
    
    async def request_data(self, link, uid, params):
        try:
            data = await http.AioSession.get(link.format(uid=uid), headers = self.headers, params=params, proxy= self.proxy)
            if data is None:
                raise StarRailCardError(4, "Failed to get data from API, please try again")
        except aiohttp.ClientConnectionError:
            raise StarRailCardError(1, "Server is not responding")
        except aiohttp.ClientResponseError as e:
            raise StarRailCardError(e.status, f"Server returned status code {e.status}")

        return data
    
    async def get_link_api(self, api_id: ApiType = ApiType.MiHoMo):
        if api_id == 1:
            if self.parsed_info:
                self.parsed_info = False
                return MiHoMoLink.PARSED
            else:
                return MiHoMoLink.NAKED
        elif api_id == 2:     
            return EnkaLink.ENKA   
    
    async def get_parsed(self, data: dict):
        if self.ua_lang:
            await ukrainization.TranslateDataManager().load_translate_data()
        
        data = await AssetEnkaParsed(data).collect()
    
    
    async def get(self, uid: Union[int,str], cookie: dict, api: ApiType = ApiType.MiHoMo):
        if not api in [1,2,3]:
            raise Exception("Invalid api value passed.")
        
        params = await self.get_params(api_id = api)
        
        link = await self.get_link_api(api = api)
            
                 
        if api == 3:
            try:
                import genshin
            except ImportError:
                raise StarRailCardError(100, "Install the genshin.py module\n- pip install genshin")
                
            player = await self.request_data(MiHoMoLink.NAKED, uid, params)
            
            try:
                client = genshin.Client(cookie, game= genshin.Game.STARRAIL, lang= self.convert_lang)
                data = await client.get_starrail_characters(uid)
            except Exception as e:
                print(e)
                raise
            #data = await self.get_parsed()
            
            #data = await AssetHoYoLabParsed(data, data.property_info).collect()

            return 
        
        else:
            data = await self.request_data(link,uid,params)

            if self.parsed_info:
                data = await self.get_parsed(data)
            
        return api_mihomo.MiHoMoApi(player=player["player"], characters= data, dont_update_link = False)
        
            

import asyncio    

async def main():
    await ShowcaseApi().get(7544644, api = 1)

asyncio.run(main())


import math
from pathlib import Path
from copy import deepcopy

from ...tools.json_data import JsonManager
from ...tools.enums import PathData

CHARACTERS_LINK_ICON = "icon/{catalog}/{character_id}.png"
CHARACTERS_LINK_IMAGE = "image/{catalog}/{character_id}.png"
ENKA_INDEX = Path(__file__).parent / "assets" / "enka_api" / "index"
ENKA = Path(__file__).parent.parent / "assets" / "enka_api"

field = {
    "BaseHP": "hp",
    "BaseAttack": "atk",
    "BaseDefence": "def",
}


class AssetEnkaParsed:
    def __init__(self, data) -> None:
        self.data = data
        self.lang = "en"
    
    async def load_assets(self):
        self.character = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "characters.json").read()
        self.element = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "elements.json").read()
        self.path =  await JsonManager(PathData.ENKA_INDEX.value / self.lang / "paths.json").read()
        self.character_rank = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "character_ranks.json").read()
        self.character_promotions = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "character_promotions.json").read()
        self.skill = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "character_skills.json").read()
        self.skill_trees_info = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "character_skill_trees.json").read()
        self.light_cone = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "light_cones.json").read()
        self.propertie = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "properties.json").read()
        self.light_cone_rank = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "light_cone_ranks.json").read()
        self.relics = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "relics.json").read()
        self.relics_set = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "relic_sets.json").read()
        self.relic_main_affixes = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "relic_main_affixes.json").read()
        self.relic_sub_affixes = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "relic_sub_affixes.json").read()
        self.relic_set = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "relic_sets.json").read()
        self.light_cone_promotion = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "light_cone_promotions.json").read()
        self.avatar = await JsonManager(PathData.ENKA_INDEX.value / self.lang / "avatars.json").read()
    
    async def collect(self, build = False):
        await self.load_assets()
        if build:
            charters = []
            for key in self.data["detailInfo"]["avatarDetailList"]:
                for k in key:
                    if isinstance(key[k], list):
                        for data in key[k]:
                            buildInfo = {
                                "id": data.get("id", 0),
                                "name_build": data.get("name", ""),
                            }
                            charters.append(await self.get_character(data["avatar_data"], buildInfo))
                    else:
                        buildInfo = {
                                "id": key[k].get("id", 0),
                                "name_build": key[k].get("name", ""),
                            }
                        charters.append(await self.get_character(key[k], buildInfo))
        else:
            charters = [await self.get_character(key) for key in self.data["detailInfo"]["avatarDetailList"]]
            
        data = {
            "player": await self.get_player(),
            "characters": charters
        }
        
        return data
    
    async def get_memory_data(self):
        data = self.data["detailInfo"]["recordInfo"]["challengeInfo"]
        return {
            "chaos_id": data.get("scheduleGroupId", 0),
            "chaos_level": data.get("scheduleMaxLevel", 0),
            "abyss_level": data.get("abyssLevel", 0),
            "abyss_star_count": data.get("abyssStarCount", 0),
        }
        pass
    
    async def get_space_info(self):
        data = self.data["detailInfo"]["recordInfo"]
        return {
            "relic_count": data["relicCount"],
            "music_count": data["musicCount"],
            "book_count": data["bookCount"],
            "universe_level": data["maxRogueChallengeScore"],
            "light_cone_count": data["equipmentCount"],
            "avatar_count": data["avatarCount"],
            "achievement_count": data["achievementCount"],
            "memory_data": await self.get_memory_data()
        }
    async def get_avatar_info(self, avatar_id):
        if str(avatar_id) not in self.avatar:
            return None
        return {
            "id": avatar_id,
            "name": self.avatar[avatar_id]["name"],
            "icon": self.avatar[avatar_id]["icon"]
        }
    async def get_player(self):

        return {
            "uid": self.data["detailInfo"]["uid"],
            "nickname": self.data["detailInfo"]["nickname"],
            "level": self.data["detailInfo"]["level"],
            "world_level": self.data["detailInfo"]["worldLevel"],
            "friend_count": self.data["detailInfo"]["friendCount"],
            "avatar": await self.get_avatar_info(str(self.data["detailInfo"]["headIcon"])),
            "signature": self.data["detailInfo"]["signature"],
            "is_display": self.data["detailInfo"]["isDisplayAvatar"],
            "space_info": await self.get_space_info()
            
            
        }
    
    async def get_light_cone(self, data):
        info = self.light_cone.get(str(data.get("tid")))
        
        id = str(data.get("tid"))
        name = info.get("name")
        rarity = info.get("rarity")
        rank = data.get("rank")
        level = data.get("level")
        promotion = data.get("promotion")
        icon = info.get("icon")
        preview = info.get("preview")
        portrait = info.get("portrait")
        path = await self.get_path(info.get("path"))
        attributes = await self.get_light_cone_attribute_from_promotion(id, promotion, level)
        properties = await self.get_light_cone_property_from_rank(id, rank)
        
        return {
            "id": id,
            "name": name,
            "rarity": rarity,
            "rank": rank,
            "level": level,
            "promotion": promotion,
            "icon": icon,
            "preview": preview,
            "portrait": portrait,
            "path": path,
            "attributes": attributes,
            "properties": properties,
        }
    
    async def get_light_cone_property_from_rank(self, id: str, rank: int):
        if id not in self.light_cone_rank:
            return []
        if rank not in range(1, 5 + 1):
            return []
        properties = []
        
        for i in self.light_cone_rank[id]["properties"][rank - 1]:
            if i["type"] not in self.propertie:
                continue
            property = self.propertie[i["type"]]
            properties.append(
                {
                    "type": i["type"],
                    "field": property["field"],
                    "name": property["name"],
                    "icon": property["icon"],
                    "value": i["value"],
                    "display": await self.get_display(i["value"], property["percent"]),
                    "percent": property["percent"],
                }
            )
        return properties
    
    async def get_light_cone_attribute_from_promotion(self, id: str, promotion: int, level: int):
        if id not in self.light_cone_promotion:
            return []
        if promotion not in range(0, 6 + 1):  # 0-6
            return []
        if level not in range(1, 80 + 1):  # 1-80
            return []
        attributes = []
        for k, v in self.light_cone_promotion[id]["values"][promotion].items():
            property = None
            for i in self.propertie.values():
                if i["field"] == k:
                    property = i
                    break
            if property is None:
                continue
            attributes.append(
                {
                    "field": k,
                    "name": property["name"],
                    "icon": property["icon"],
                    "value": v["base"] + v["step"] * (level - 1),
                    "display": await self.get_display(v["base"] + v["step"] * (level - 1), property["percent"]),
                    "percent": property["percent"],
                }
            )
        return attributes
    
    async def get_character(self, data, build = {}):
        id = str(data.get("avatarId"))
        name = self.character.get(str(id)).get("name")
        rarity = self.character.get(str(id)).get("rarity")
        rank = data.get("rank", 0)
        level = data.get("level")
        promotion = data.get("promotion", 0)
        icon = self.character.get(str(id)).get("icon")
        preview = self.character.get(str(id)).get("preview")
        portrait = self.character.get(str(id)).get("portrait")
        rank_icons = [await self.get_rank_icons(key) for key in self.character.get(str(id)).get("ranks")]
        path = await self.get_path(self.character.get(str(id)).get("path"))
        element = await self.get_element(self.character.get(str(id)).get("element"))
        skills = await self.get_skill(data, str(id), self.character.get(str(id)).get("element"))
        skill_trees = await self.get_skill_trees(data, str(id), rank)
        light_cone = await self.get_light_cone(data.get("equipment")) if data.get("equipment") else None
        attributes = await self.get_attributes(str(id), promotion, level)
        properties = await self.get_properties(str(id), data["skillTreeList"])
                
        relic_infos = [await self.get_relic_info(relic) for relic in data["relicList"]]
        relics = [
            relic_info for relic_info in relic_infos if relic_info is not None
        ]
        relic_sets = await self.get_relic_sets_info(relics) if relics else []
        
        attributes = await self.merge_attribute(
            [
                attributes,
                light_cone["attributes"] if light_cone else [],
            ]
        )

        
        relic_properties = []
        for relic in relics:
            if relic["main_affix"]:
                relic_properties.append(relic["main_affix"])
            relic_properties += [
                {
                    "type": affix["type"],
                    "field": affix["field"],
                    "name": affix["name"],
                    "icon": affix["icon"],
                    "value": affix["value"],
                    "display": affix["display"],
                    "percent": affix["percent"],
                } for affix in relic["sub_affix"]
            ]

        for relic_set in relic_sets:
            relic_properties += relic_set["properties"]
                
        properties = await self.merge_property(
            [
                properties,
                light_cone["properties"] if light_cone else [],
                relic_properties,
            ]
        )
        
        additions = await self.get_additions(attributes, properties)        
      
        return {
            "id": id,
            "name": name,
            "rarity": rarity,
            "rank": rank,
            "level": level,
            "promotion": promotion,
            "icon": icon,
            "preview": preview,
            "portrait": portrait,
            "rank_icons": rank_icons,
            "path": path,
            "element": element,
            "skills": skills,
            "skill_trees": skill_trees,
            "light_cone": light_cone,
            
            "relics": relics,
            "relic_sets": relic_sets,
            "attributes": attributes,
            "additions": additions,
            "properties": properties,
            "build": build,
            "pos": [data.get("pos", 0)],
        }
        
    async def get_additions(self, attributes, properties):
        attribute_dict = {}
        addition_dict = {}
        for attribute in attributes:
            if not attribute["field"]  in addition_dict:
                attribute_dict[attribute["field"]] = attribute["value"]
        for property in properties:
            if (
                self.propertie[property["type"]]["ratio"]
                and property["field"] in attribute_dict
            ):
                value = property["value"] * attribute_dict[property["field"]]
            else:
                value = property["value"]
            if property["field"] not in addition_dict:
                addition_dict[property["field"]] = value
            else:
                addition_dict[property["field"]] += value
        additions = []
        for k, v in addition_dict.items():
            property = None
            for i in self.propertie.values():
                if i["field"] == k:
                    property = i
                    break
                
            if property is None:
                continue
            
            additions.append(
                {
                    "field": k,
                    "name": property["name"],
                    "icon": property["icon"],
                    "value": v,
                    "display": await self.get_display(v, property["percent"]),
                    "percent": property["percent"],
                }
            )
            
        return additions
    
    async def get_attributes(self, character_id, promotion, level):
        
        if character_id not in self.character_promotions:
            return []
        if promotion not in range(0, 6 + 1):  # 0-6
            return []
        if level not in range(1, 80 + 1):  # 1-80
            return []
        attributes = []
        for k, v in self.character_promotions[character_id]["values"][promotion].items():
            property = None
            for i in self.propertie.values():
                if i["field"] == k:
                    property = i
                    break
            if property is None:
                continue
            
            attributes.append(
                {
                    "field": k,
                    "name": property["name"],
                    "icon": property["icon"],
                    "value": v["base"] + v["step"] * (level - 1),
                    "display": await self.get_display(v["base"] + v["step"] * (level - 1), property["percent"]),
                    "percent": property["percent"]
                }
            )
        return attributes
    
    async def get_properties(self, characters_id, info):
        skill_trees = self.character[characters_id]["skill_trees"]
        properties = []
        
        for skill_tree in info:
            pointId = str(skill_tree["pointId"])
            if pointId in skill_trees and pointId in self.skill_trees_info:
                property_list = (self.skill_trees_info[pointId]["levels"][skill_tree["level"] - 1]["properties"])
                for i in property_list:
                    if self.propertie.get(i.get("type")) is None or i["value"] <= 0:
                        continue
                    property = self.propertie[i["type"]]             
                    properties.append(
                        {
                            "type": i["type"],
                            "field": property["field"],
                            "name": property["name"],
                            "icon": property["icon"],
                            "value": i["value"],
                            "display": await self.get_display( i["value"], property["percent"]),
                            "percent": property["percent"],
                            "type": i["type"],
                        }
                    )
        return properties
    
    async def get_skill_trees(self, data, character_id, rank):
        return [{
            "id": str(key["pointId"]),
            "level": key["level"],
            "anchor": self.skill_trees_info.get(str(key["pointId"]))["anchor"],
            "icon": self.skill_trees_info.get(str(key["pointId"]))["icon"],
            "max_level": await self.get_max_level(self.character.get(character_id)["ranks"], str(key["pointId"]), rank),
            "parent": await self.get_parent(key["pointId"])
            } for key in data["skillTreeList"]]
    
    async def get_skill(self, data, character_data, element):

        character_data = self.character.get(character_data)
        
        response= []
        
        for keys in data["skillTreeList"]:
            pointId = str(keys["pointId"])
            if self.skill_trees_info.get(pointId)["level_up_skills"] == []:
                continue
            
            skills_info_id = str(self.skill_trees_info.get(pointId)["level_up_skills"][0]["id"])
            
            if not str(skills_info_id) in character_data["skills"]:
                continue
            
            info = self.skill.get(skills_info_id)
            
            max_level = self.skill_trees_info.get(pointId)["max_level"]
            if keys["level"] > max_level:
                max_level = info["max_level"]

            response.append(
                {
                    "id": skills_info_id,
                    "name": info["name"],
                    "level": keys["level"],
                    "max_level": max_level,
                    "element": await self.get_element(element),
                    "type": info["type"],
                    "type_text": info["type_text"],
                    "effect": info["effect"],
                    "effect_text": info["effect_text"],
                    "simple_desc": info["simple_desc"],
                    "desc": info["desc"],
                    "icon": info["icon"],
                }
            )
            
        return response
                
    async def get_relic_info(self, data):
        if str(data["tid"]) not in self.relics:
            return None
        
        relics_info = self.relics.get(str(data["tid"]))
        
        info = {
                "id": str(data["tid"]),
                "name": relics_info["name"],
                "set_id": str(data["_flat"]["setID"]),
                "set_name": self.relics_set[relics_info["set_id"]]["name"],
                "rarity": relics_info["rarity"],
                "level": data["level"],
                "icon": relics_info['icon'],
                "main_affix": await self.get_relic_main_affix(str(data["tid"]), data["level"], str(data["mainAffixId"])),
                "sub_affix": await self.get_relic_sub_affix(str(data["tid"]), data["subAffixList"])
                }

        return info
    
    async def get_relic_sub_affix(self, relict_id, sub_affix_info):
        if relict_id not in self.relics:
            return []
        
        sub_affix_group = self.relics[relict_id]["sub_affix_id"]
        if sub_affix_group not in self.relic_sub_affixes:
            return []
        
        properties = []
        for sub_affix in sub_affix_info:
            if str(sub_affix["affixId"]) not in self.relic_sub_affixes[sub_affix_group]["affixes"]:
                continue
            affix = self.relic_sub_affixes[sub_affix_group]["affixes"][str(sub_affix["affixId"]) ]

            property = self.propertie[affix["property"]]
            
            properties.append(
                {
                    "type": affix["property"],
                    "field": property["field"],
                    "name": property["name"],
                    "icon": property["icon"],
                    "value": affix["base"] * sub_affix["cnt"] + affix["step"] * sub_affix.get("step", 0),
                    "display":  await self.get_display(affix["base"] * sub_affix["cnt"] + affix["step"] * sub_affix.get("step", 0), property["percent"]),
                    "percent": property["percent"],
                    "count": sub_affix["cnt"],
                    "step": sub_affix.get("step", 0)
                }
            )
        return properties
    
    async def get_relic_main_affix(self, relict_id, level, main_affix_id):
        if str(relict_id) not in self.relics:
            return None
        if not main_affix_id:
            return None
        
        main_affix_group = self.relics[str(relict_id)]["main_affix_id"]            
            
        if main_affix_group not in self.relic_main_affixes or main_affix_id not in self.relic_main_affixes[main_affix_group]["affixes"]:
            return None
        
        affix = self.relic_main_affixes[main_affix_group]["affixes"][main_affix_id]
        
        property = self.propertie[affix["property"]]
          
        return {
                "type": affix["property"],
                "field": property["field"],
                "name": property["name"],
                "icon": property["icon"],
                "value":affix["base"] + + affix["step"] * level,
                "display": await self.get_display(affix["base"] + + affix["step"] * level, property["percent"]),
                "percent": property["percent"]
            }  
    
    async def get_relic_sets_info(self, data):
        set_num = {}
        for relic in data:
            if relic["set_id"] not in set_num:
                set_num[relic["set_id"]] = 1
            else:
                set_num[relic["set_id"]] += 1
                
        relic_sets = []
        for k, v in set_num.items():
            info = self.relic_set[str(k)]
            if v >= 2:
                prop = [{
                    "type": i["type"],
                    "field": self.propertie.get(i["type"])["field"],
                    "name": self.propertie.get(i["type"])["name"],
                    "icon": self.propertie.get(i["type"])["icon"],
                    "value": i["value"],
                    "display": await self.get_display(i["value"], self.propertie.get(i["type"])["percent"]),
                    "percent": self.propertie.get(i["type"])["percent"],
                } for i in info["properties"][0]]
                
                relic_sets.append(
                    {
                        "id": k,
                        "name": info["name"],
                        "icon": info["icon"],
                        "num": 2,
                        "desc": info["desc"][0],
                        "properties": prop,
                    }
                )
            if v >= 4:
                prop = [
                    {
                        "type": i["type"],
                        "field": self.propertie.get(i["type"])["field"],
                        "name": self.propertie.get(i["type"])["name"],
                        "icon": self.propertie.get(i["type"])["icon"],
                        "value": i["value"],
                        "display": await self.get_display(i["value"], self.propertie.get(i["type"])["percent"]),
                        "percent": self.propertie.get(i["type"])["percent"],
                    } for i in info["properties"][1] if len(info["properties"]) > 1
                ]

                relic_sets.append(
                    {
                        "id": k,
                        "name": info["name"],
                        "icon": info["icon"],
                        "num": 4,
                        "desc": info["desc"][1] if len(info["desc"]) > 1 else "",
                        "properties": prop,
                    }
                )
        return relic_sets
    
    async def merge_property(self, properties):
        property_dict = {}
        for property_list in properties:
            for property in property_list:
                if isinstance(property, list):
                    if len(property) == 0:
                        continue
                    property = property[0]

                if property["type"] not in property_dict:
                    property_dict[property["type"]] = {}
                    property_dict[property["type"]]["value"] = property["value"]
                    property_dict[property["type"]]["origin"] = deepcopy(property)
                else:
                    property_dict[property["type"]]["value"] += property["value"]
        property_res = []
        
        for v in property_dict.values():
            
            info = v["origin"]
            info["value"] = v["value"]
            info["display"] = await self.get_display(v["value"], info["percent"])

            property_res.append(info)
        return property_res
    
    async def merge_attribute(self, attributes):
        attribute_dict = {}
        for attribute_list in attributes:
            for attribute in attribute_list:
                if attribute["field"] not in attribute_dict:
                    attribute_dict[attribute["field"]] = {}
                    attribute_dict[attribute["field"]]["value"] = attribute["value"] 
                    attribute_dict[attribute["field"]]["origin"] = deepcopy(attribute)
                else:
                    attribute_dict[attribute["field"]]["value"] += attribute["value"] 
        attribute_res = []
        for v in attribute_dict.values():
            attribute_info = v["origin"]
            attribute_info["value"] = v["value"]
            attribute_info["display"] = await self.get_display(
                v["value"], attribute_info["percent"]
            )
            attribute_res.append(attribute_info)
            
        return attribute_res
    
    async def get_parent(self, skill_id):
        parent = self.skill_trees_info.get(str(skill_id)).get("pre_points")
        if parent == []:
            return None
        else:
            return parent[0]   
         
    async def get_rank_icons(self, rank_id):
        return self.character_rank.get(rank_id).get("icon")
    
    async def get_path(self, path):
        return {"id": self.path.get(path)["id"], "name": self.path.get(path)["name"], "icon": self.path.get(path)["icon"]}

    async def get_element(self, element):
        return {"id": self.element.get(element)["id"], "name": self.element.get(element)["name"], "color": self.element.get(element)["color"], "icon": self.element.get(element)["icon"]}
    
    async def get_max_level(self, ranks_list,skill_id,rank):
        
        skill_trees_info = self.skill_trees_info.get(skill_id)
        
        max_level = skill_trees_info["max_level"]
        if skill_trees_info.get("level_up_skills", []) != []:
            if int(skill_trees_info.get("level_up_skills")[0]["id"]) in ranks_list[:rank]:
                max_level = self.skill[skill_trees_info.get("level_up_skills")[0]["id"]]["max_level"]

        return max_level
    
    async def get_display(self, value, percent):
        if percent:
            return format(math.floor(value * 1000) / 10.0, ".1f") + "%"
        else:
            return f"{math.floor(value)}"


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

import anyio

import io
from PIL import ImageDraw,Image,ImageChops, ImageSequence, ImageFilter
import functools

from ..tools import git, options, treePaths
from ..tools.calculator import stats
from ..model.style import Card

from ..tools import pill

_of = git.ImageCache()
_of.set_mapping(4)


_DEFAULT_SCORE = {'count': 0, 
                  'rolls': {}, 
                  'rank': {'name': 'N/A', 
                           'color': (255, 255, 255, 255)
                    }
}

class Create:
    def __init__(self, data, lang, art, hide_uid, uid, seeleland,remove_logo, color) -> None:
        self.remove_logo = remove_logo
        self.data = data
        self.lang = lang
        self.art = art 
        self.hide_uid = hide_uid
        self.uid = uid
        
        self.gif = False
        self.GIFT_BG = []
        
        self.seeleland = seeleland
        self.total_eff = 0
        if not color is None:
            self.element_color = color
        else:
            self.element_color = self.data.element.color.rgba
    
    
    
    async def create_background(self):
        self.background = Image.new(Card.RGBA, Card.background_size, self.element_color)
        image_background =   Image.new(Card.RGBA, Card.background_size, (0,0,0,0))
        background = Image.new(Card.RGBA, Card.background_size, (0,0,0,0))
        
        logo = await _of.LOGO_GIT_INV
        shadow = await _of.background_shadow
        background_overlay = await _of.background_overlay
        
        if self.art:
            user_image = await pill.get_center_size(Card.art_size, self.art)
            position_art = Card.position_art
            mask = await _of.background_maska_art
        else:
            user_image = await pill.get_center_scale(Card.splash_size, await pill.get_download_img(self.data.portrait))
            position_art = Card.position_splash_art
            mask = await _of.background_maska_art #await _of.background_maska

        ll = await pill.light_level(self.element_color)

        if ll > 0.59 or self.element_color == (0, 255, 156, 255):
            background_dark = Image.new(Card.RGBA, Card.background_size, (0,0,0,70))  
            self.background.alpha_composite(background_dark)
            
        image_background.alpha_composite(user_image, position_art)
        self.background = ImageChops.soft_light(self.background, background_overlay.convert(Card.RGBA))

        self.background.alpha_composite(shadow,Card.position_shadow)
        background.paste(image_background.convert(Card.RGBA),(0,0), mask.convert("L"))
        self.background.alpha_composite(background)

        if not self.remove_logo:
            self.background.alpha_composite(logo,(1004,0))
        
        '''if not self.hide_uid:
            d = ImageDraw.Draw(self.background)
            font_10 = await pill.get_font(Card.font_uid)
            d.text(Card.position_uid, f"UID: {self.uid}", font=font_10, fill=(255, 255, 255, 255))
            d.text((Card.position_uid[0]-1, Card.position_uid[1]), f"UID: {self.uid}", font=font_10, fill=(0, 0, 0, 0))
        '''

    async def create_light_cone(self):
        if self.data.light_cone is None:
            self.light_cone_background = Image.new("RGBA", (1, 1), (0, 0, 0, 0))
            return 
        
        mask = await _of.maska_light_cones
        frame_light_cones = await _of.frame_light_cones
        blic_light_cones = await _of.blic_light_cones
        
        light_cone_holst = Image.new(Card.RGBA, Card.lc_background_size, (0, 0, 0, 0))
        light_cone_holst_image = Image.new(Card.RGBA, Card.lc_background_size, (0, 0, 0, 0))

        image = await pill.get_download_img(self.data.light_cone.portrait, size=Card.lc_image_size)
        light_cone_holst_image.alpha_composite(image, Card.lc_image_position)

        light_cone_holst.paste(light_cone_holst_image, (0, 0), mask.convert("L"))

        light_cone_holst_image = Image.new(Card.RGBA, Card.lc_background_size, (0, 0, 0, 0))
        
        shadow, frame, self.color = await pill.get_resource_light_cones(self.data.light_cone.rarity)
        
        light_cone_holst_image.alpha_composite(shadow)
        light_cone_holst_image.alpha_composite(frame, Card.lc_frame_position)
        light_cone_holst_image.alpha_composite(light_cone_holst)
        light_cone_holst_image.alpha_composite(frame_light_cones)
        light_cone_holst_image.alpha_composite(blic_light_cones)
        light_cone_holst_image.alpha_composite(frame)

        
        background_stat = await _of.stats_light_cones
        background_stat = background_stat.copy()
        font_size = 29
        font_color = (193, 162, 103, 255)

        font = await pill.get_font(font_size)

        d = ImageDraw.Draw(background_stat)
        rank = options.ups(self.data.light_cone.rank)
        x = int(font.getlength(rank) / 2)
        d.text((17 - x, 2), rank, font=font, fill=font_color)

        font = await pill.get_font(19)

        max_level = options.max_lvl(self.data.light_cone.promotion)

        d.text((40,9), f"{self.lang.lvl}: {self.data.light_cone.level}/{max_level}", font=font, fill=(255, 255, 255, 255))
        
        d.text((49, 53), self.data.light_cone.attributes[0].display, font=font, fill=(255, 255, 255, 255)) #HP
        d.text((152, 53), self.data.light_cone.attributes[1].display, font=font, fill=(255, 255, 255, 255)) #ATK
        d.text((49, 100), self.data.light_cone.attributes[2].display, font=font, fill=(255, 255, 255, 255)) #DEF
        
        stars = await options.get_stars(self.data.light_cone.rarity,type=3)
        
        name_light_cone = await pill.create_image_with_text(self.data.light_cone.name, 24, max_width=230, color=(255, 255, 255, 255))
        line = Image.new("RGBA", (1, name_light_cone.size[1] + 2), self.color)

        self.light_cone_background = Image.new("RGBA", (447, 255), (0, 0, 0, 0))
        self.light_cone_background.alpha_composite(light_cone_holst_image.resize((183, 244)))
        self.light_cone_background.alpha_composite(stars,(0,6))
        self.light_cone_background.alpha_composite(line, (191, 19))
        self.light_cone_background.alpha_composite(name_light_cone, (200, 23))
        
        y = int(22 + line.size[1])

        self.light_cone_background.alpha_composite(background_stat, (188, y))
    
    async def create_stats(self):
        self.background_stats = Image.new(Card.RGBA, (502, 657), (0, 0, 0, 0))

        combined_attributes = {}
        dop = {}
        
        for attribute in self.data.attributes + self.data.additions:
            field = attribute.field
            if field in combined_attributes:
                combined_attributes[field].value += attribute.value
                dop[field]["dop"] = f"+{attribute.display}"
            else:
                dop[field] = {"main": attribute.display, "dop": 0}
                combined_attributes[field] = attribute

        dop = {key: value for key, value in dop.items() if value['dop'] != 0}
        
        result = []
        
        self.element_color = await pill.get_light_pixel_color(self.element_color[:3], up= True)
        
        async for i, background in pill.create_stats(combined_attributes, dop, 
                      Card.stat_font_dop,Card.stat_font,
                      Card.stat_line_size, Card.stat_name_size, Card.stat_max_width, Card.stat_icon_size,
                      Card.stat_icon_position, Card.stat_name_position, Card.stat_x_position,
                      self.element_color,Card.stat_value_font
                      ,Card.stat_y_no_dop,Card.stat_y_yes_dop,Card.stat_y_dop,Card.stat_line_add):
            result.append(background)
                
        if len(result) <= 1:
            x = 0 
        
        x = int(657 / (len(result)))
        position_line = 0
        for key in result:
            self.background_stats.alpha_composite(key, (0, position_line))
            position_line += x
    
    async def get_score(self):
        self.score_info = await stats.Calculator(self.data).start(self.hoyo)
        
    async def create_relict(self,relict):
        background_main = Image.new(Card.RGBA,Card.relict_size, (0,0,0,0))
        background = Image.new(Card.RGBA,Card.relict_size, (0,0,0,0))
        background_image = Image.new("RGBA",Card.relict_size, (0,0,0,0))
        relict_frame = await _of.relict_frame
        relict_mask = await _of.relict_maska
        relict_score_frame = await _of.relict_score_frame
        relict_line = await _of.relict_line
        background_main.alpha_composite(relict_frame)
        
        image = await pill.get_download_img(relict.icon, size= Card.relict_icon_size)
        background_image.alpha_composite(image,Card.relict_icon_position)
        background.paste(background_image,(0,0),relict_mask.convert("L"))
        background_main.alpha_composite(relict_score_frame,(8,0))
        background_main.alpha_composite(background)
        background_main.alpha_composite(relict_line)
        score = self.score_info["score"].get(str(relict.id),_DEFAULT_SCORE)
        main_stat_icon = await pill.get_download_img(relict.main_affix.icon, size= Card.relict_main_stat_icon_size)
        color = options.color_element.get(relict.main_affix.type, None)
        if not color is None:
            main_stat_icon = await pill.recolor_image(main_stat_icon, color[:3])
        stars = await options.get_stars(relict.rarity, type = 1)
        background_main.alpha_composite(stars,Card.relict_position_star)
        background_main.alpha_composite(main_stat_icon,Card.relict_main_stat_icon_position)
        font_22 = await pill.get_font(22)
        font_14 = await pill.get_font(14)
        font_18 = await pill.get_font(18)
        
        d = ImageDraw.Draw(background_main)
        x = 135 - int(font_22.getlength(relict.main_affix.display))
        d.text((x, 86), relict.main_affix.display, font= font_22, fill= (255,255,255,255))
        
        x = 121 - int(font_22.getlength(f"+{relict.level}")/2)
        d.text((x, 110), f"+{relict.level}", font= font_14, fill= self.element_color)
        y_icon = 21
        y_roll = 30
        y_text = 27
                
        eff = 0
        for k in relict.sub_affix:
            icon = await pill.get_download_img(k.icon, size=Card.relict_sub_icon_size)
            background_main.alpha_composite(icon,(142,y_icon))
            scoreR = score["rolls"].get(k.type,0)
            if k.type in self.score_info["bad"]:
                d.text((173, y_text), str(k.display), font=font_18, fill= (255,255,255,255))
            else:
                eff += 1
                d.text((173, y_text), str(k.display), font=font_18, fill= options.color_scoreR.get(scoreR,(255,255,255,255)))
            d.text((232, y_roll), f"+{scoreR}", font=font_18, fill= options.color_scoreR.get(scoreR,(255,255,255,255)))
            y_icon += 27
            y_text += 28
            y_roll += 27
            
        d.text((22, 4), "Score:", font=font_14, fill = (255,255,255,255))
        d.text((70, 4), str(score["count"]), font=font_14, fill = score["rank"]["color"])
        
        d.text((105, 4), "Rank:", font=font_14, fill = (255,255,255,255))
        d.text((149, 4), str(score["rank"]["name"]), font=font_14, fill = score["rank"]["color"])
        
        d.text((187, 4), "Eff Stat:", font=font_14, fill = (255,255,255,255))
        d.text((247, 4), str(eff), font=font_14, fill = options.color_scoreR.get(eff,(255,255,255,255)))
        
        self.total_eff += eff
        
        line_f = await _of.relict_frame_line
        line_f = line_f.copy()
        line_f = await pill.recolor_image(line_f, self.element_color[:3])
        background_main.alpha_composite(line_f,(0,21))
        
        return {"position": str(relict.id)[-1:] , "img": background_main}
    
    async def create_score_total(self):
        self.background_score = Image.new("RGBA",(536,43), (0,0,0,0))
        
        sclor_bg = await _of.relict_backgroundl_score_line
        sclor_bg = sclor_bg.copy().resize((536,16))
        sclor_bg_color = await _of.relict_full_score_line
        sclor_bg_color = sclor_bg_color.copy().resize((536,16))
        sclor_bg_color = await pill.recolor_image(sclor_bg_color, self.element_color[:3])
        
        result_percentage = (self.score_info["total_score"]['count'] / 260.53) * 100
        pixels_to_fill = int((result_percentage / 100)*536)
        if pixels_to_fill <= 0:
            pixels_to_fill = 1
        sclor_bg.alpha_composite(sclor_bg_color.resize((pixels_to_fill,16)))
        
        font_18 = await pill.get_font(14)
        
        d = ImageDraw.Draw(sclor_bg)
        x = 257 - int(font_18.getlength(f"{round(result_percentage,1)}%")/2)
        frame = await pill.get_colors(await pill.crop_image(sclor_bg), 15, common=True, radius=5, quality=800) 
        ll = await pill.light_level(frame)
        if ll > 0.6 or frame == (51, 51, 51):
            d.text((x, 0), f"{round(result_percentage,1)}%", font=font_18, fill = (0,0,0,166))
        else:
            d.text((x, 0), f"{round(result_percentage,1)}%", font=font_18, fill = (255,255,255,255))
        
        self.background_score.alpha_composite(sclor_bg,(0,27))
        
        font_21 = await pill.get_font(15)
        d = ImageDraw.Draw(self.background_score)
        
        d.text((249, -2), "Score:", font=font_21, fill = (255,255,255,255))
        d.text((300, -2), str(self.score_info["total_score"]['count']), font=font_21, fill = self.score_info["total_score"]["rank"]["color"])
        
        d.text((17, -2), "Summary Rank:", font=font_21, fill = (255,255,255,255))
        d.text((140, -2), str(self.score_info["total_score"]["rank"]["name"]), font=font_21, fill = self.score_info["total_score"]["rank"]["color"])
        
        d.text((393, -2), "Eff Stat:", font=font_21, fill = (255,255,255,255))
        d.text((457, -2), str(self.total_eff), font=font_21, fill = options.color_scoreR.get(self.total_eff,(255,255,255,255)))
    
    
    async def create_relict_sets(self):
        rel_set = options.calculator_relict_sets(self.data.relic_sets)

        self.background_sets = Image.new(Card.RGBA, Card.sets_background, (0,0,0,0))
        
        font = await pill.get_font(Card.sets_font)
        
        line_items = []
        i = 0
        for key in rel_set:
            sets = rel_set[key]
            holst_line = Image.new(Card.RGBA, Card.sets_line_size, (0,0,0,0))
            background_count = await _of.relict_count_sets
            background_count = background_count.copy()
            d = ImageDraw.Draw(background_count)
            d.text(Card.sets_count_position, str(sets["num"]), font=font, fill= Card.sets_name_color)
            d = ImageDraw.Draw(holst_line)
            sets_name_font,size = await pill.get_text_size_frame(sets["name"],15,492)
            if key[:1] == "1":
                holst_line.alpha_composite(background_count)
                d.text((39, 4), sets["name"], font=sets_name_font, fill=Card.sets_name_color)
                line_items.append({"setap": i, "line": holst_line})
                i += 1
            else:
                holst_line.alpha_composite(background_count,(503 ,0))
                d.text((int(490-size), 4), sets["name"], font=sets_name_font, fill=Card.sets_name_color)
                line_items.append({"setap": 1, "line": holst_line})
                
        for key in line_items:
            if key["setap"] == 1:
                self.background_sets.alpha_composite(key["line"],(0,29))
            elif key["setap"] == 2:
                self.background_sets.alpha_composite(key["line"],(0,29))
            else:
                self.background_sets.alpha_composite(key["line"],(0,0))
    
    async def build_relict(self):
        self.background_relict = Image.new(Card.RGBA, (536,577), (0,0,0,0))
        none_relict = await _of.none_relict
        position = {
            "1": (0,55),
            "2": (274,55),
            "3": (0,205),
            "4": (274,205),
            "5": (0,347),
            "6": (274,347),
        }
        
        map = {"1": None, "2": None, "3": None, "4": None, "5": None, "6": None}
        
        for key in self.relict:
            self.background_relict.alpha_composite(key["img"].resize((262,128)), position.get(key["position"]))
            map[key["position"]] = 0
        
        for key in map:
            if map[key] is None:
                self.background_relict.alpha_composite(none_relict, position.get(key))

        self.background_relict.alpha_composite(self.background_score,(0,0))
        self.background_relict.alpha_composite(self.background_sets,(2,486))
    
    async def build(self):
        self.background.alpha_composite(self.light_cone_background.resize(Card.build_lc_size),Card.build_lc_position)
        self.background.alpha_composite(self.background_stats, Card.build_stats_position)
        self.background.alpha_composite(self.background_relict, Card.build_relict_position)
        self.background.alpha_composite(self.stats, Card.build_skill_position)
        self.background.alpha_composite(self.background_skills,Card.build_constant_position)
        self.background.alpha_composite(self.background_name, (23,4))
    
    async def create_skill(self):
        holst_main = Image.new("RGBA", (69, 296), (0, 0, 0, 0))
        holst_dop = Image.new("RGBA", (265, 293), (0, 0, 0, 0))
        line = Image.new("RGBA", (15, 2), (255, 255, 255, 255))
        holst_point_add = Image.new("RGBA", (145, 68), (0, 0, 0, 0))
        point_add_position = 0
        position_dop = treePaths.tree.get(self.data.path.id)
        position_main_y = 0

        for key in self.data.skill_trees:
            if key.max_level != 1:
                font = await pill.get_font(24)
                color = (227, 215, 194, 255)
                icon_skill = Image.new("RGBA", (69, 68), (0, 0, 0, 0))
                background = await _of.talants_background
                background = background.copy()
                icon = await pill.get_download_img(key.icon, size=(55, 55))
                
                frame = await _of.talants_adaptationt_frame
                if key.max_level > 10:
                    color = options.color_element_talant.get(self.data.element.id.lower(), (255,255,255,255))
                if key.anchor in ["Point19", "Point20"]:
                    icon = await pill.recolor_image(icon, (215, 194, 227))
                    frame = await pill.recolor_image(frame.copy(),  (215, 194, 227))
                else:
                    icon = await pill.recolor_image(icon, (227, 215, 194))
                    frame = await pill.recolor_image(frame.copy(), (color[0], color[1], color[2]))
                
                background.alpha_composite(icon, (4, 4))
                count_background = await _of.talants_count
                count_background = count_background.copy()
                d = ImageDraw.Draw(count_background)
                x = int(font.getlength(str(key.level)) / 2)
                d.text((15 - x, 4), str(key.level), font=font, fill=color)
                background.alpha_composite(frame)
                icon_skill.alpha_composite(background, (5, 0))
                icon_skill.alpha_composite(count_background, (0, 37))
                if key.anchor in ["Point19", "Point20"]:
                    holst_point_add.alpha_composite(icon_skill, (point_add_position, 0))
                    point_add_position += 78

                else:
                    holst_main.alpha_composite(icon_skill, (0, position_main_y))
                    position_main_y += 76
            else:
                if key.anchor in ["Point05", "Point06", "Point07", "Point08"]:
                    icon_background = await _of.talants_main_stats
                    icon_background = icon_background.copy()
                    icon = await pill.get_download_img(key.icon, size=(40, 40))
                    icon = await pill.recolor_image(icon, (0, 0, 0))
                    icon_background.alpha_composite(icon, (13, 14))
                    if key.level == 0:
                        icon_background = await pill.apply_opacity(icon_background, opacity=0.5)
                    holst_dop.alpha_composite(icon_background, treePaths.position_point[key.anchor])
                else:
                    icon_background = await _of.talants_mini_stats
                    icon_background = icon_background.copy()
                    icon = await pill.get_download_img(key.icon, size=(40, 40))
                    icon = await pill.recolor_image(icon, (0, 0, 0))
                    icon_background.alpha_composite(icon, (0, 0))
                    if key.level == 0:
                        icon_background = await pill.apply_opacity(icon_background, opacity=0.5)
                    holst_dop.alpha_composite(icon_background, (222 - position_dop[str(key.id)[-2:]][0],position_dop[str(key.id)[-2:]][1]))
                    
                    holst_dop.alpha_composite(line, (266 - position_dop[str(key.id)[-2:]][0], position_dop[str(key.id)[-2:]][1] + 21))
        
        
        self.stats = Image.new("RGBA", (345, 367), (0, 0, 0, 0))        
        self.stats.alpha_composite(holst_point_add, (120, 4))
        self.stats.alpha_composite(holst_main, (275, 72))
        self.stats.alpha_composite(holst_dop, (0, 72))
        
    
    
    async def create_constant(self):
        self.background_skills = Image.new(Card.RGBA, Card.constant_size_background, (0, 0, 0, 0))
        y = 0
        rank = self.data.rank
        for key in self.data.rank_icons[:rank]:
            icon = await pill.get_download_img(key, size=Card.constant_size_icon)
            self.background_skills.alpha_composite(icon,(0,y))
            y += 46

        for key in self.data.rank_icons[rank:]:
            icon = await pill.get_download_img(key, size= Card.constant_size_icon)
            icon = await pill.apply_opacity(icon, opacity=Card.constant_size_icon_opacity)
            self.background_skills.alpha_composite(icon,(0,y))
            y += 46
        
    async def create_name(self):
        self.background_name = Image.new(Card.RGBA, Card.name_size, (0, 0, 0, 0))
        d = ImageDraw.Draw(self.background_name)
        names = await pill.create_image_with_text(self.data.name, Card.font_name_size , max_width=Card.name_with, color= Card.name_color)
        names_dark = await pill.recolor_image(names,(0,0,0))
        
        if names.size[1] <= Card.name_h_max:
            self.background_name.alpha_composite(names_dark,(Card.name_position-1,int(Card.name_h_max-names.size[1]/2)+1))
            self.background_name.alpha_composite(names,(Card.name_position,int(Card.name_h_max-names.size[1]/2)))
        else:
            self.background_name.alpha_composite(names_dark,(Card.name_position-1,int(Card.name_h_min-names.size[1]/2)+1))
            self.background_name.alpha_composite(names,(Card.name_position,int(Card.name_h_max-names.size[1]/2)))
        
        font_17 = await pill.get_font(Card.font_name_level)
        max_level = options.max_lvl(self.data.promotion)
        level = f"{self.lang.lvl}: {self.data.level}/{max_level}"
        d.text((Card.position_name_level[0]-1,Card.position_name_level[1]+1,), level, font=font_17, fill= (0,0,0,255))
        d.text(Card.position_name_level, level, font=font_17, fill= Card.color_name_level)
        starts = await options.get_stars(self.data.rarity)
        self.background_name.alpha_composite(starts, Card.position_name_star)

    async def start(self, build = None, hoyo = False ):
        _of.set_mapping(4)
        self.hoyo = hoyo
        
        if self.art:
            if "gif" in self.art:
                self.gif = True
            self.art = await pill.get_user_image(self.art)
            if self.gif:    
                n = 2           
                frame_count = 0
                with io.BytesIO(self.art) as f:
                    self.art = Image.open(f)
                    for frame in ImageSequence.Iterator(self.art):
                        self.element_color = await pill.get_colors(frame.convert("RGBA"), 15, common=True, radius=5, quality=800)
                        if frame_count >= 50:
                            break
                        if frame_count % n != 0:
                            self.art = frame.convert("RGBA")
                            await self.create_background()
                            self.GIFT_BG.append(self.background.copy())
                        frame_count += 1
            else:
                self.element_color = await pill.get_colors(self.art, 15, common=True, radius=5, quality=800)
                await self.create_background()
        else:
            await self.create_background()
        
        async with anyio.create_task_group() as tasks:
            tasks.start_soon(self.create_light_cone) #
            tasks.start_soon(self.create_stats) #
            tasks.start_soon(self.create_name) #
            tasks.start_soon(self.create_constant) #
            tasks.start_soon(self.create_relict_sets) #
            tasks.start_soon(self.get_score) #
            tasks.start_soon(self.create_skill)
            
        async def wait_all(*funcs):
            results = [None] * len(funcs)
            
            async with anyio.create_task_group() as tasks:
                async def process(func, i):
                    results[i] = await func()
                
                for i, func in enumerate(funcs):
                    tasks.start_soon(process, func, i)
            
            return results
        
        self.relict = await wait_all(*[
            functools.partial(self.create_relict, key)
            for key in self.data.relics
        ])
        
        await self.create_score_total()
        
        await self.build_relict()
        
        await self.build()
        
        
        data = {
                "id": self.data.id,
                "name": self.data.name,
                "animation": self.gif,
                "rarity": self.data.rarity,
                "card": self.background,
                "size": Card.background_size,
                "color": self.element_color,
                "build": build
            }
            
        return data


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

import anyio
from typing import Union
import traceback

from .src.tools import cache, http, ukrainization, options, translator, git
from .src.generator import style_relict_score, style_ticket, style_profile_phone, style_card
from .src.api import api, enka, error, hoyolab
from .src.model import StarRailCard,api_mihomo
from .src.tools.pill import image_control


class HoYoCard:
    def __init__(self, cookie: dict, lang: str = "en", character_art = None, character_id = None, seeleland: bool = False,
                 user_font = None, save: bool = False, asset_save: bool = False, boost_speed: bool = False, remove_logo: bool = False,
                 cache = {"maxsize": 150, "ttl": 300}, proxy: str =  None, color: dict = None):
        
        self.lang = lang
        self.character_art  = character_art
        self.character_id  = character_id
        self.seeleland = seeleland
        self.user_font = user_font
        self.save = save
        self.asset_save = asset_save
        self.remove_logo = remove_logo
        self.cache = cache
        self.boost_speed = boost_speed
        self.proxy = proxy
        self.color = color
        self.cookie = cookie
        
    async def __aenter__(self):
        cache.Cache.get_cache(maxsize = self.cache.get("maxsize", 150), ttl = self.cache.get("ttl", 300))
        await http.AioSession.enter(self.proxy)
        
        await git.ImageCache.set_assets_download(self.asset_save)
        
        if self.character_id:
            self.character_id = await options.get_charter_id(self.character_id)
        
        if self.character_art:
            if not isinstance(self.character_art, dict):
                raise TypeError(4,"The character_art parameter must be a dictionary, where the key is the name of the character, and the parameter is an image.\nExample: character_art = {'1235': 'img.png', '1235': ['img.png','http.../img2.png']} or {'123596': 'img.png', '123854': 'http.../img2.png', ...}")
            else:
                self.character_art = await options.get_character_art(self.character_art)
        
        if not self.lang in translator.SUPPORTED_LANGUAGES:
            self.lang = "en"
        
        if self.lang == "ua":
            await ukrainization.TranslateDataManager().check_update()
        
        self.translateLang = translator.Translator(self.lang)

        if self.user_font:
            await git.change_font(font_path = self.user_font)
        
        image_control._boost_speed = self.boost_speed

        
        if isinstance(self.color, dict):
            self.color = await options.get_color_user(self.color)
        
        if self.remove_logo:
            print("""
                Thank you for using our StarRailCard!
                By removing the GitHub logo from the generated results, you acknowledge supporting the author through one of the following links:
                - Patreon: https://www.patreon.com/deviantapi/membership
                - Ko-fi: https://ko-fi.com/dezzso

                Failure to contribute to the project may lead to the author discontinuing updates in the future.
                """)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await http.AioSession.exit(exc_type, exc, tb)
        
    async def set_lang(self, lang):
        """Sets the language

        Args:
            lang (str): lang (str): Language in which generation will take place, supported languages: cht, cn, de, en, es, fr, id, jp, kr, pt, ru, th, vi, ua . Defaults to "en".
        """
        if not lang in translator.SUPPORTED_LANGUAGES:
            lang = "en"
            self.lang = "en"
        else:
            self.lang = lang
        
        if lang == "ua":
            await ukrainization.TranslateDataManager().check_update()
        
        self.translateLang = translator.Translator(lang)
                
    async def set_user_font(self, user_font: str):
        """Will install a custom font

        Args:
            user_font (srt): Font path or font name.
        """
        await git.change_font(font_path = user_font)
    
    async def create(self, uid: Union[int,str], style: int = 1, hide_uid: bool = False):
        """Function for generating character cards

        Args:
            uid (int): UID of the user in the game
            style (int, optional): Card style. Defaults to 1.
            hide_uid (bool, optional): Hide UID. Defaults to False.

        Returns:
            StarRail: A class object containing profile information and a character cards
        """        
        try:
            data = await hoyolab.HoYoLabApi(uid, lang= self.lang).get(self.cookie)
        except Exception as e:
            print(e)
            raise error.StarRailCardError(8, "To use the HoYoLab API you need to download/update the asset")
  
        result = []
        style, style_settings = await options.style_setting(style, None)
                
        try:
            player = data.player.model_dump()
        except:
            player = data.player
        
        response = {
            "settings": {
                "uid": int(uid),
                "lang": self.lang,
                "hide_uid": hide_uid,
                "save": self.save,
                "force_update": False,
                "style": int(style)
            },
            "player": player,
            "card": None,
            "character_name": [],
            "character_id": [],
        }
                
        async with anyio.create_task_group() as tasks:
            
            for key in data.characters:
                async def get_result(key):    
                    try:               
                        response["character_id"].append(key.id)
                        response["character_name"].append(key.name)
                        
                        if self.character_id:
                            if not str(key.id) in self.character_id:
                                return  
                        
                        if self.color:
                            color = self.color.get(str(key.id))
                        else:
                            color = None
                        
                        art = None
                        if self.character_art:
                            if str(key.id) in self.character_art:
                                art = self.character_art[str(key.id)]
                        if style == 1:
                            result.append(await style_relict_score.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start(hoyo = True))
                        elif style == 2:
                            result.append(await style_ticket.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start(hoyo = True))
                        elif style == 3:
                            result.append(await style_card.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start(hoyo = True))
                    except Exception as e:
                        print(f"Error in get_result for character {key.id}: {e}")
                        
                tasks.start_soon(get_result, key)
                    
        response["card"] = result
        
        if self.lang == "ua":
            StarRailCard.UA_LANG = True
        else:
            StarRailCard.UA_LANG = False
        
        if self.save:
            async with anyio.create_task_group() as tasks:
                for key in response["card"]:
                    if key["animation"]:
                        continue
                    tasks.start_soon(options.save_card,uid,key["card"],key["id"])        
                
        return StarRailCard.StarRail(**response)


class Card:
    def __init__(self, lang: str = "en", character_art = None, character_id = None, seeleland: bool = False,
                 user_font = None, save: bool = False, asset_save: bool = False, boost_speed: bool = False, remove_logo: bool = False,
                 cache = {"maxsize": 150, "ttl": 300}, enka: bool = False, api_data: api_mihomo.MiHoMoApi = None, proxy: str =  None, 
                 user_agent: str = None, color: dict = None):
        """Main class for generating cards

        Args:
            lang (str, optional): Language in which generation will take place, supported languages: cht, cn, de, en, es, fr, id, jp, kr, pt, ru, th, vi, ua . Defaults to "en".
            character_art (dict, optional): A dictionary that contains a key - character id and link value (If you want to pass several images for 1 character, you can use list in the values). Defaults to None.
            character_id (str, optional): List character ids separated by commas. Defaults to None.
            seeleland (bool, optional): Enable Seeleland statistics. Defaults to False.
            user_font (srt, optional): Font path or font name. Defaults to None.
            save (bool, optional): Whether to save images or not (Does not work for animated cards). Defaults to False.
            asset_save (bool, optional): Save assets to the device so that in subsequent calls you do not have to download them, but open them from the device. Defaults to False.
            boost_speed (bool, optional): Allows you to download generation resources to your device for further use without downloading. !!!Fills up the device memory!!!.
            remove_logo (bool, optional): Remove GItHub logo. Defaults to False.
            cashe (dict, optional): Set your cache settings. Defaults to {"maxsize": 150, "ttl": 300}.
            api_data (MiHoMoApi, optional): Pass your data received via: api.ApiMiHoMo(uid,"en").get()
            proxy (str, optional): Proxy as a string: http://111.111.111.111:8888
            user_agent (str, optional): Custom User-Agent.
            color: (dict), Dictionary: {"character_id_1": (255,255,255,255), "character_id_2": (0,0,0,255),...}.
        """
        self.lang = lang
        self.character_art  = character_art
        self.character_id  = character_id
        self.seeleland = seeleland
        self.user_font = user_font
        self.save = save
        self.asset_save = asset_save
        self.remove_logo = remove_logo
        self.cache = cache
        self.enka = enka
        self.boost_speed = boost_speed
        self.api_data = api_data
        self.proxy = proxy
        self.user_agent = options.get_user_agent(user_agent)
        self.color = color

        
    async def __aenter__(self):
        cache.Cache.get_cache(maxsize = self.cache.get("maxsize", 150), ttl = self.cache.get("ttl", 300))
        await http.AioSession.enter(self.proxy)
        
        await git.ImageCache.set_assets_download(self.asset_save)
        
        if self.character_id:
            self.character_id = await options.get_charter_id(self.character_id)
        
        if self.character_art:
            if not isinstance(self.character_art, dict):
                raise TypeError(4,"The character_art parameter must be a dictionary, where the key is the name of the character, and the parameter is an image.\nExample: character_art = {'1235': 'img.png', '1235': ['img.png','http.../img2.png']} or {'123596': 'img.png', '123854': 'http.../img2.png', ...}")
            else:
                self.character_art = await options.get_character_art(self.character_art)
        
        if not self.lang in translator.SUPPORTED_LANGUAGES:
            self.lang = "en"
        
        if self.lang == "ua":
            await ukrainization.TranslateDataManager().check_update()
        
        self.translateLang = translator.Translator(self.lang)

        if self.user_font:
            await git.change_font(font_path = self.user_font)
        
        image_control._boost_speed = self.boost_speed

        
        if isinstance(self.color, dict):
            self.color = await options.get_color_user(self.color)
        
        if self.remove_logo:
            print("""
                Thank you for using our StarRailCard!
                By removing the GitHub logo from the generated results, you acknowledge supporting the author through one of the following links:
                - Patreon: https://www.patreon.com/deviantapi/membership
                - Ko-fi: https://ko-fi.com/dezzso

                Failure to contribute to the project may lead to the author discontinuing updates in the future.
                """)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await http.AioSession.exit(exc_type, exc, tb)
        
    async def set_lang(self, lang):
        """Sets the language

        Args:
            lang (str): lang (str): Language in which generation will take place, supported languages: cht, cn, de, en, es, fr, id, jp, kr, pt, ru, th, vi, ua . Defaults to "en".
        """
        if not lang in translator.SUPPORTED_LANGUAGES:
            lang = "en"
            self.lang = "en"
        else:
            self.lang = lang
        
        if lang == "ua":
            await ukrainization.TranslateDataManager().check_update()
        
        self.translateLang = translator.Translator(lang)
                
    async def set_user_font(self, user_font: str):
        """Will install a custom font

        Args:
            user_font (srt): Font path or font name.
        """
        await git.change_font(font_path = user_font)
        
    async def create_profile(self, uid: Union[int,str], style: bool = 1, hide_uid: bool = False, background = None, force_update: bool = False):
        """Function for generating a user profile card

        Args:
            uid (Union[int,str]): UID of the user in the game.
            style (int, optional): Card style. Defaults to 1.
            hide_uid (bool, optional): Hide UID. Defaults to False.
            background (str, optional): Link to custom card background. Defaults to None.
            force_update (bool, optional): Forced update of user data. Defaults to False.

        Returns:
            StarRail: A class object containing profile information and a profile card
        """
        if self.api_data is None:
            if self.enka:
                try:
                    data = await enka.ApiEnkaNetwork(uid, lang= self.lang).get()
                except Exception as e:
                    print(e)
                    print("To use the EnkaNetwork API you need to download/update the asset\nExample: await enka.ApiEnkaNetwork().update_assets()")
                    data = await api.ApiMiHoMo(uid, lang= self.lang, force_update = force_update, user_agent= self.user_agent).get()
            else:
                data = await api.ApiMiHoMo(uid, lang= self.lang, force_update = force_update, user_agent= self.user_agent).get()
        else:
            data = self.api_data
            
        try:
            player = data.player.model_dump()
        except:
            player = data.player
            
        response = {
            "settings": {
                "uid": int(uid),
                "lang": self.lang,
                "hide_uid": hide_uid,
                "save": self.save,
                "force_update": force_update,
                "style": int(style)
            },
            "player": player,
            "card": None,
            "character_name": [],
            "character_id": [],
        }
        
        if not str(style) in ["1"]:
            style = 1
            
        if style == 1:
            response["card"] = await style_profile_phone.Create(data,self.translateLang,self.character_art,hide_uid,uid,background, self.remove_logo).start()

        for key in data.characters:
            response["character_id"].append(key.id)
            response["character_name"].append(key.name)
        
        if self.save:
            await options.save_card(uid,response["card"],"profile")
        
        return StarRailCard.StarRail(**response)
        
    async def create(self, uid: Union[int,str], style: int = 1, hide_uid: bool = False, force_update: bool = False, style_settings = None):
        """Function for generating character cards

        Args:
            uid (int): UID of the user in the game
            style (int, optional): Card style. Defaults to 1.
            hide_uid (bool, optional): Hide UID. Defaults to False.
            force_update (bool, optional): forced update of user data. Defaults to False.
            style_settings (dict, optional): not implemented yet. Defaults to None.

        Returns:
            StarRail: A class object containing profile information and a character cards
        """
        
        if self.api_data is None:
            if self.enka:
                try:
                    data = await enka.ApiEnkaNetwork(uid, lang= self.lang).get()
                except Exception as e:
                    print(e)
                    print("To use the EnkaNetwork API you need to download/update the asset")
                    data = await api.ApiMiHoMo(uid, lang= self.lang, force_update = force_update, user_agent= self.user_agent).get()
            else:
                data = await api.ApiMiHoMo(uid, lang= self.lang, force_update = force_update, user_agent= self.user_agent).get()
        else:
            data = self.api_data

        result = []
        
        style, style_settings = await options.style_setting(style, style_settings)
                
        try:
            player = data.player.model_dump()
        except:
            player = data.player
        
        response = {
            "settings": {
                "uid": int(uid),
                "lang": self.lang,
                "hide_uid": hide_uid,
                "save": self.save,
                "force_update": force_update,
                "style": int(style)
            },
            "player": player,
            "card": None,
            "character_name": [],
            "character_id": [],
        }
                
        async with anyio.create_task_group() as tasks:
            
            for key in data.characters:
                async def get_result(key):    
                    try:               
                        response["character_id"].append(key.id)
                        response["character_name"].append(key.name)
                        
                        if self.character_id:
                            if not str(key.id) in self.character_id:
                                return  
                        
                        if self.color:
                            color = self.color.get(str(key.id))
                        else:
                            color = None
                        
                        art = None
                        if self.character_art:
                            if str(key.id) in self.character_art:
                                art = self.character_art[str(key.id)]
                        if style == 1:
                            result.append(await style_relict_score.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start())
                        elif style == 2:
                            result.append(await style_ticket.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start())
                        elif style == 3:
                            result.append(await style_card.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start())
                    except Exception as e:
                        print(f"Error in get_result for character {key.id}:\n{traceback.format_exc()}")
                        print(f"Error in get_result for character {key.id}: {e}")
                        
                tasks.start_soon(get_result, key)
                    
        response["card"] = result
        
        if self.lang == "ua":
            StarRailCard.UA_LANG = True
        else:
            StarRailCard.UA_LANG = False
        
        if self.save:
            async with anyio.create_task_group() as tasks:
                for key in response["card"]:
                    if key["animation"]:
                        continue
                    tasks.start_soon(options.save_card,uid,key["card"],key["id"])        
                                
        return StarRailCard.StarRail(**response)
    
    
    async def create_build(self, uid: Union[int,str], name: str, hash: str, style: int = 1, hide_uid: bool = False):
        """Function for generating character cards

        Args:
            uid (int): UID of the user in the game
            name (str): NickName EnkaNetwork
            hash (str): hash obtained using the method: get_hash()
            style (int, optional): Card style. Defaults to 1.
            hide_uid (bool, optional): Hide UID. Defaults to False.
            style_settings (dict, optional): not implemented yet. Defaults to None.

        Returns:
            StarRail: A class object containing profile information and a character cards
        """
        
        if self.api_data is None:
            try:
                data = await enka.ApiEnkaNetwork().get_build(name, hash, uid)
            except Exception as e:
                print(e)
                raise error.StarRailCardError(8, "To use the HoYoLab API you need to download/update the asset")
        else:
            data = self.api_data

        result = []
        
        style, style_settings = await options.style_setting(style, None)
                
        try:
            player = data.player.model_dump()
        except:
            player = data.player
        
        response = {
            "settings": {
                "uid": int(uid),
                "lang": self.lang,
                "hide_uid": hide_uid,
                "save": self.save,
                "force_update": False,
                "style": int(style)
            },
            "player": player,
            "card": None,
            "character_name": [],
            "character_id": [],
        }
                
        async with anyio.create_task_group() as tasks:
            
            for key in data.characters:
                async def get_result(key):
                    try:               
                        response["character_id"].append(key.id)
                        response["character_name"].append(key.name)
                        
                        if self.character_id:
                            if not str(key.id) in self.character_id:
                                return  
                        
                        if self.color:
                            color = self.color.get(str(key.id))
                        else:
                            color = None
                        
                        art = None
                        if self.character_art:
                            if str(key.id) in self.character_art:
                                art = self.character_art[str(key.id)]
                            if str(key.build["id"]) in self.character_art:
                                art = self.character_art[str(key.build["id"])]
                        if style == 1:
                            result.append(await style_relict_score.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start(build= key.build))
                        elif style == 2:
                            result.append(await style_ticket.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start(build= key.build))
                        elif style == 3:
                            result.append(await style_card.Create(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo, color).start(build= key.build))
                    except Exception as e:
                        print(f"Error in get_result for character {key.id}: {e}")
                        
                tasks.start_soon(get_result, key)
                    
        response["card"] = result
        
        if self.lang == "ua":
            StarRailCard.UA_LANG = True
        else:
            StarRailCard.UA_LANG = False
        
        if self.save:
            async with anyio.create_task_group() as tasks:
                for key in response["card"]:
                    print(key)
                    if key["animation"]:
                        continue
                    tasks.start_soon(options.save_card,uid,key["card"],f'{key["id"]}_{key["build"]["name_build"]}')        
                                
        return StarRailCard.StarRail(**response)